/// ///////////////////////////////////////////////////////////////////////////////////////// ///
///							  Nonlinear Model Predictive Control(NMPC)						          ///
///							  Created by:	Torben Falleth Olsen							                ///
///                       Edited by: ...                                                      ///
/// ///////////////////////////////////////////////////////////////////////////////////////// ///


#include "NMPC.h"
#include "Eigen/Dense"
#include <iostream>
#include <time.h>       /* time */
#include <stdlib.h>     /* srand, rand */
#include <cmath>
#include <casadi/casadi.hpp>
#include <sfh/text.h>
#include <chrono>
using namespace casadi;

namespace AquaRobotics
{


	NMPC::NMPC(const string& simObjectName, ISimObjectCreator* const creator) : SimObject(simObjectName)
	{
      creator->AddInport("VehiclePosition", 6, &m_InVehiclePos);
      creator->AddInport("VehicleVelocity", 6, &m_InVehicleVel);
      creator->AddInport("VehicleAcceleration", 6, &m_InVehicleAcc);
      creator->AddInport("NextWaypoint", 3, &m_InWaypoint);
      //creator->AddInport("MovingObstacle", 3, &m_InObstaclePos);

      // Setup output ports
      creator->AddOutport("NextNode", 3, PORT_FUNCTION(NMPC::OutNextNode)); // Creates an outport with name "Out" and dimension 3. The function OutFunction calculates the output value.
      creator->GetIntParam("NumberOfNodesToLog", &m_NodesToLog, 50);
      for (int i = 0; i < m_NodesToLog; i++) {
         creator->AddOutportIndexed("Path" + sfh::text::Int2Str(i + 1), 3, INDEXED_PORT_FUNCTION(NMPC::OutNode), i);
         m_OutNodes.push_back(vec3(0, 0, 0));
      }

      creator->GetIntParam("ObstacleNumber", &m_ObstacleNumber);
      creator->GetIntParam("NumberOfWaypoints", &m_WaypointNumber);

      // Waypoints
      for (int i = 1; i <= m_WaypointNumber; i++) {
         vec3 WP;
         creator->AddOutportIndexed("Waypoint" + std::to_string(i), 3, INDEXED_PORT_FUNCTION(NMPC::OutWaypoint), i);
         creator->GetDoubleParamArray("Waypoint" + std::to_string(i), WP.data(), 3);
         m_Waypoints.push_back(WP);
      }

      // Obstacle

      std::srand(time(NULL)); // Creates a random number sequence function, which is accsessed though the "rand()" function 

      m_InObstaclePos = new ISignalPort*[m_ObstacleNumber];
      m_InObstacleVel = new ISignalPort*[m_ObstacleNumber];

      for (int i = 0; i < m_ObstacleNumber; i++) {
         creator->AddInport("Obstacle" + std::to_string(i) + "Position", 3, &m_InObstaclePos[i]);
         creator->AddInport("Obstacle" + std::to_string(i) + "Velocity", 3, &m_InObstacleVel[i]);
         creator->GetDoubleParam("Obstacle" + std::to_string(i) + "Radius", &m_ObstacleRad);
         obs_radius_init.push_back(m_ObstacleRad);
      }
      

		///////////////////////////////////////////////////////////////////////////
		//									    Parameters											 //
		///////////////////////////////////////////////////////////////////////////
      //
	   N                     = 10;                        // - Prediction Horizon
      N_min                 = 2;                         // - Minimum update horizon from Goal 
      Update_frequency      = 10.0;                      // [Hz] - Solver frequency
      Vehicle_radius        = 0.5;                       // [m] - Vehicle radius
      Vehicle_mass          = 90.0;                      // [kg] - Vehicle mass
      seg_len               = 1.0;                       // [m] - The length between path segments
      observation_dist      = 0.5;                       // [m] - The observation distance
      //
      //
      //////////////////////////////////////////////////////////////////////
      ///////// -- OBS: These need to correspont appropriately ! -- ////////
      //////////////////////////////////////////////////////////////////////
      // 
      // --- Extra distance to obstacles. NEEDS TO MATCH AMOUNT OF OBSTACLES ! --- //
      Obstacle_dist = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0}; // [m] - Extra distance between obstacle and vehicle
      //
      // --- Points of interests (for MPC). NEEDS TO MATCH AMOUNT OF GOALS (i.e Goal.size = 2, then PointsOfInterest = {{},{}} ) --- //
      PointsOfInterest = {{}};
      //
      // --- Points of interests. NEEDS TO BE EQUAL TO HAVE THE SAME POINTS AS THE ONE ABOVE (i.e PointsOfInterest = {{vertcat(MX(22.0), MX(5.0), MX(0.0))}} , PoiOfInt = {{vec3(22.0, 5.0, 0.0)}} ) --- //
      PoiOfInt = {{}};
      //
      ///////////////////////////////////////////////////////////////////////////
      ///////////////////////////////////////////////////////////////////////////
      ///////////////////////////////////////////////////////////////////////////


      // --- Creates the initial path --- //
      InitializePath();

      // --- Limits the prediction horizon from becoming larger that the path --- //
      if (N > path[wp_ind].size()-1) 
      {
         N = path[wp_ind].size()-1;
      }

      // --- Stores the initial prediction horizon --- //
      N_init = N;


      // --- Creates break time out of the frequency --- //
      break_time = (1 / Update_frequency) * CLOCKS_PER_SEC;

      // --- Sets the optimal path equal to the initial path --- //
      m_Path = path[wp_ind];
	}

   NMPC::~NMPC() 
   {
      delete[] m_InObstaclePos;
      delete[] m_InObstacleVel;
   }



   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							         Initialization of Path             						          ///
   ///							                                                 		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   void NMPC::InitializePath()
   { 
      for (int k = 1; k < m_WaypointNumber; k++) {

         // --- Calculates the distance between two points in x,y,z --- //
         double distx = (m_Waypoints[k][0] - m_Waypoints[k - 1][0]);                // Path length along x-axis
         double disty = (m_Waypoints[k][1] - m_Waypoints[k - 1][1]);                // Path length along y-axis
         double distz = (m_Waypoints[k][2] - m_Waypoints[k - 1][2]);                // Path length along z-axis

         // --- Calculates the amount of nodes in the path --- //
         n = int((1 / seg_len) * (m_Waypoints[k] - m_Waypoints[k - 1]).norm());     // Path nodes

         // --- Calculates the distance between each node in x,y,z --- //
         double segx = distx / (n-1);                                               // Path length along x-axis
         double segy = disty / (n-1);                                               // Path length along y-axis
         double segz = distz / (n-1);                                               // Path length along z-axis

         path_wp.clear();

         // --- Creates a straight line path with 'n' nodes --- //
         for (int i = 0; i < n; i++) {
            vec3 node = vec3(m_Waypoints[k - 1][0] + (i * segx), m_Waypoints[k - 1][1] + (i * segy), m_Waypoints[k - 1][2] + (i * segz));

            // --- Creates the path --- //
            path_wp.push_back(node);
         }

         // --- Stores path --- //
         path.push_back(path_wp);

      }  

   }

   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							               Updates the Initial Path     						          ///
   ///							                     					            		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   void NMPC::UpdateInitialPath()
   {
      // --- If the vehicle is further away than the minimum prediction horizon --- //
      if (ind < path[wp_ind].size() - N_min) {

         // --- Calculates the distance between the current index and the goal in x,y,z --- //
         double distx = (m_Waypoints[wp_ind+1][0] - m_Path[ind][0]);                // Path length along x-axis
         double disty = (m_Waypoints[wp_ind+1][1] - m_Path[ind][1]);                // Path length along y-axis
         double distz = (m_Waypoints[wp_ind+1][2] - m_Path[ind][2]);                // Path length along z-axis

         // --- Calculates the amount of nodes in the path --- //
         n = path[wp_ind].size() - ind;                                             // Path nodes

         // --- Calculates the distance between each node in x,y,z --- //
         double segx = distx / (n - 1);                                             // Path length along x-axis
         double segy = disty / (n - 1);                                             // Path length along y-axis
         double segz = distz / (n - 1);                                             // Path length along z-axis

         // --- Creates a straight line path with 'n' nodes ---//
         for (int i = 0; i < n; i++) {
            vec3 node         = vec3(m_Path[ind][0] + (i * segx), m_Path[ind][1] + (i * segy), m_Path[ind][2] + (i * segz));

            // --- Updates the initial path --- //
            path[wp_ind].at(i + ind) = node;
         }

      }
   }

   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							               Updates the Path             						          ///
   ///							                     					            		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   void NMPC::UpdatePath(vec3 NED)
   {

      // --- If the vehicle is closer than 0.3m to a node --- //
      if ((NED - m_Path[ind]).norm() < 0.3) 
      {

         // If N is increased to a higher number that N_init,
         // then it will decrease with one every time the vehicle hits a new node
         if (N > N_init) {
            N = N - 1;
         }

         // Updates the current path index
         ind++;


         // --- Deletes point of interest when the vehicle is within region --- //
         for (int i = 0; i < PoiOfInt[wp_ind].size(); i++) 
         {
            if ((NED - PoiOfInt[wp_ind][i]).norm() <= observation_dist + Vehicle_radius) 
            {
               PoiOfInt[wp_ind].erase(PoiOfInt[wp_ind].begin() + i);
               PointsOfInterest[wp_ind].erase(PointsOfInterest[wp_ind].begin() + i);
            }
         }


         // --- If the vehicle is close to the goal --- //
         if ((NED - m_Waypoints[wp_ind + 1]).norm() < 0.5 && wp_ind < path.size()-1) 
         {

            wp_ind++;                        // Update the goal index
            m_Path = path[wp_ind];           // Update optimal path to be equal to the new initial path
            N = N_init;                      // Reset the prediction horizon to the initial prediction horizon
            ind    = 0;                      // Reset the path index to zero 
            count  = 1;                      // Allow the mpc to run on the next iteration
         }

         count = 1;                          // Allow the mpc to run on the next iteration

         // --- Recalculate the prediction horizon --- //
         Calc_N();

         // --- Update the number of path nodes --- //
         UpdatePathNodes();

         // --- Update the initial path --- //
         UpdateInitialPath();

         // --- Update the optimal path to be equal to the initial path --- //
         m_Path = path[wp_ind];
      }

      // Updating the converging node
      m_NextNode = m_Path[ind]; 

   }



   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							        Updates the nodes of the Path       						          ///
   ///							                     					            		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   void NMPC::UpdatePathNodes()
   {
      // --- If the vehicle is further away than the minimum prediction horizon --- //
      if (ind < path[wp_ind].size() - N_min) 
      {
         // --- Calculates the length between initial path nodes --- //
         double distx = (path[wp_ind][N - 1 + ind][0] - path[wp_ind][ind][0]);         
         double disty = (path[wp_ind][N - 1 + ind][1] - path[wp_ind][ind][1]);         
         double distz = (path[wp_ind][N - 1 + ind][2] - path[wp_ind][ind][2]);

         // --- Store the previous path index --- //
         path_prev_index = path[wp_ind].size();

         
         double l_f = 0;
         double l_s = 0;

         // --- Calculate the arc length of the initial path and the optimal path --- //
         for (int i = ind; i < ind+N-1; i++) 
         {

            l_f += (m_Path[i + 1] - m_Path[i]).norm();
            l_s += (path[wp_ind][i + 1] - path[wp_ind][i]).norm();
         }

         // --- Calculate the new amount of nodes --- //
         n_new = int(l_f/seg_len);

         // --- Calculates the distance between each node in x,y,z --- //
         double       segx_prime = distx / (n_new); // Path length along x-axis
         double       segy_prime = disty / (n_new); // Path length along y-axis
         double       segz_prime = distz / (n_new); // Path length along z-axis


         // --- Erase the previous horizon nodes --- //
         for (int i = N-1; i > 0; i--) 
         {
            path[wp_ind].erase(path[wp_ind].begin() + i + ind);
         }

         // --- Create and insert the new nodes --- //
         for (int i = 0; i < n_new; i++) {
            vec3 node = vec3(path[wp_ind][ind][0] + (i * segx_prime), path[wp_ind][ind][1] + (i * segy_prime), path[wp_ind][ind][2] + (i * segz_prime));
            path[wp_ind].insert(path[wp_ind].begin() + i + ind + 1, node);
         }

         // --- Get the new path index --- //
         path_index = path[wp_ind].size();

         // --- Calculate the increase or decrease in nodes --- //
         diff = path_index - path_prev_index;

         // --- Update the prediction horizon --- //
         if ((N + diff) >= N_init) {

            N = N + diff;

         }

      }
      
   }



   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							      Recalculate the Prediction horizon    						          ///
   ///							                     					            		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   void NMPC::Calc_N()
   {
      int Index = 0;

      
      for (int i = ind + N - 1; i < path[wp_ind].size(); i++) {
         int coun = 0;

         // --- Checks to see if node N is in collision --- //
         for (int j = 0; j < m_Obs.size(); j++) {
            if ((path[wp_ind][i] - m_Obs[j]).norm() <= (obs_radius[j] + Vehicle_radius + Obstacle_dist[j])) {
               coun++;
            }
         }
         // --- If node N collides with obstacle, add index --- //
         if (coun > 0) {
            Index++;

         } else {
            break;
         }
      }

      // --- Update the horizon dependent on collisions --- //
      N = N + Index;

      // If the prediction horizon preceedes the path index,
      // decreese the prediction horizon.
      if (N + ind > path[wp_ind].size()-1) {
         N = (path[wp_ind].size() - ind-1);
      }

      
   }


   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							  Creating a static obstacle from a dynamic 						          ///
   ///							                                                  		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   void NMPC::Dynamic2Static(vec3 NED)
   {
      if (ind < m_Path.size() - N_min) {



            
         StateTime.clear();
         double time_state = 0.0;

         // --- Connect the states of the path to time relative to the vehicle speed --- //
         StateTime.push_back(time_state);
         for (int i = 0; i < N - 1; i++) {
            time_state += ((m_Path[i + 1 + ind] - m_Path[i + ind]).norm() / VehicleVelocity.norm());
            StateTime.push_back(time_state);
         }

         // --- Create prediction trajectory for the obstacles --- //
         m_Obs_time_vec_all.clear();
         for (int j = 0; j < m_ObstacleNumber; j++) 
         {
            m_Obs_time_vec.clear();
            for (int i = 0; i < N-1; i++) 
            {
               m_Obs_time_vec.push_back(m_Obs[j] + (m_Obs_speed[j] * StateTime[i]));
            }
            m_Obs_time_vec_all.push_back(m_Obs_time_vec);
         }


         m_Obs_collision.clear();
         collision_index.clear();
         collision = false;

         // --- Check where the vehicle collides with the obstacle --- //
         for (int j = 0; j < m_ObstacleNumber; j++) {

            m_Obs_collision_points.clear();
            inObs = 0;
            if (m_Obs_speed[j] != vec3(0.0, 0.0, 0.0)) {
               for (int i = 0; i < N - 1; i++) {
                  vec3 path2obs = (m_Path[i + ind] - m_Obs_time_vec_all[j][i]);

                  if (path2obs.norm() <= (obs_radius[j] + Vehicle_radius + Obstacle_dist[j])) {

                     m_Obs_collision_points.push_back(m_Obs_time_vec_all[j][i]);
                     if (inObs == 0) {
                        collision_index.push_back(j);
                        inObs = 1;
                     }
                     collision = true;
                  }
               }
               m_Obs_collision.push_back(m_Obs_collision_points);
            }
            
         }

         // --- Create static obstacles --- //
         if (collision == true) 
         {
            collisions = m_Obs_collision;
         }

         // --- Update Obstacle list --- //
         for (int j = 0; j < collisions.size(); j++) {
            for (int i = 0; i < collisions[j].size(); i++) {

               obs.push_back(vertcat(MX(collisions[j][i][0]), MX(collisions[j][i][1]), MX(collisions[j][i][2])));
               m_Obs.push_back(vec3(collisions[j][i][0], collisions[j][i][1], collisions[j][i][2]));
               obs_radius.push_back(obs_radius_init[collision_index[j]]);
            }
         }
         
      }
   }


   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							             Object Function               						          ///
   ///							                  f(x)					            		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///

   MX NMPC::ObjectFunction(MX x, MX y, MX z)
   { 
      MX f = 0;
      for (int i = 0; i < (N - 1); i++)
      {
         auto vec  = vertcat(x(i), y(i), z(i));                         // Creates vector x_{i}
         auto vec1 = vertcat(x(i + 1), y(i + 1), z(i + 1));             // Creates vector x_{i+1}

         f = f + pow(norm_2(vec1 - vec), 2);                            // ||x_{i+1} - x_{i}||^2
      }
      return f;
   }


   void NMPC::UpdatePositions(const double T, const double* const X)
   {
      ROVPos = m_InVehiclePos->GetPortValue(T, X);                                  // Get vehicle position
      const double* ROVVel = m_InVehicleVel->GetPortValue(T, X);                    // Get vehicle velocity
      const double* ROVAcc = m_InVehicleAcc->GetPortValue(T, X);                    // Get vehicle Acceleration

      VehicleForce = Vehicle_mass * vec3(ROVAcc[0], ROVAcc[1], ROVAcc[2]);

      VehicleVelocity = vec3(ROVVel[0], ROVVel[1], ROVVel[2]);
      MX   VehiVelocity    = vertcat(MX(ROVVel[0]), MX(ROVVel[1]), MX(ROVVel[2]));

      obs.clear();
      m_Obs.clear();
      m_Obs_speed.clear();
      for (int i = 0; i < m_ObstacleNumber; i++) {

         const double* Obs     = m_InObstaclePos[i]->GetPortValue(T, X);            // Get Obstacle positions
         const double* Obs_velocity = m_InObstacleVel[i]->GetPortValue(T, X);       // Get Obstacle velocity

         obs.push_back(vertcat(MX(Obs[0]), MX(Obs[1]), MX(Obs[2])));
         m_Obs.push_back(vec3(Obs[0], Obs[1], Obs[2]));
         m_Obs_speed.push_back(vec3(Obs_velocity[0], Obs_velocity[1], Obs_velocity[2]));
      }
      obs_radius = obs_radius_init;
   }



	void NMPC::PreOdeFcn(const double T, const double* const X, IStateUpdater* updater)
	{

	}



   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
   ///							                    NMPC                   						          ///
   ///							                     					            		                ///
   /// ///////////////////////////////////////////////////////////////////////////////////////// ///
	void NMPC::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep)
	{
      //m_FhSimLogger->WriteLog("test", fhsim::LogLevel_Warning);    // Write to logg
      
      // --- Updates Positions and velocities --- //
      UpdatePositions(T, X);
      

      vec3 NED = vec3(ROVPos[0], ROVPos[1], ROVPos[2]);
      

      // --- Predicts collisions with dynamic obstacles --- //  
      Dynamic2Static(NED);


      // --- Limits the MPC to a frequenz and stops the replanning when at minimum planning distance --- //       
      if (((clock() - now) >= break_time || count == 1) && (ind < path[wp_ind].size() - (1 + N_min))) {
         
         count = 0;

         // --- Recalculates the prediction horizon --- //
         Calc_N();

         // --- CasADi Opti stack class --- //
         Opti opti = Opti();


         // --- Define variables --- //
         auto x = opti.variable(N);
         auto y = opti.variable(N);
         auto z = opti.variable(N);
         
         ////////////////////////////////////////////////////
         //-------------- Object Function -----------------//
         ////////////////////////////////////////////////////
         MX f = ObjectFunction(x, y, z);
         opti.minimize(f);


         ////////////////////////////////////////////////////
         //----------- Equality constraints ---------------//
         ////////////////////////////////////////////////////

         //--- c1 ---:  x_{0} = x_{start}
         opti.subject_to(x(0) == m_Path[ind][0]);
         opti.subject_to(y(0) == m_Path[ind][1]);
         opti.subject_to(z(0) == m_Path[ind][2]);



         //--- c2 ---:  x_{N} = x_{segment N}
         opti.subject_to(x(N - 1) == path[wp_ind][N - 1 + ind][0]);
         opti.subject_to(y(N - 1) == path[wp_ind][N - 1 + ind][1]);
         opti.subject_to(z(N - 1) == path[wp_ind][N - 1 + ind][2]);


         //--- c3 ---: ...

         
         ////////////////////////////////////////////////////
         //----------- Inequality constraints -------------//
         ////////////////////////////////////////////////////
         
         for (int i = 0; i < N; i++) {
            auto vec = vertcat(x(i), y(i), z(i));

            for (int k = 0; k < obs.size(); k++) {


               //--- c1 ---: ||x_{i} - x_{obs}|| > Radius_{vehi} + Radius_{Obs}
               opti.subject_to(norm_2(vec - (obs[k])) > (obs_radius[k] + Vehicle_radius + Obstacle_dist[k]));


               //--- c2 ---: || (x_{i+1} + x_{i})/2 - x_{obs})|| > Radius_{vehi} + Radius_{Obs}
               if (i < N - 1) 
               {
                  auto vec1    = vertcat(x(i + 1), y(i + 1), z(i + 1));
                  auto new_vec = 0.5*(vec + vec1);

                  opti.subject_to(norm_2(new_vec - (obs[k])) > (obs_radius[k] + Vehicle_radius + Obstacle_dist[k]));
               }

            }

         }

         //--- c3 ---: vehicle_radius < min{||x_{i} - x_{point}||} <= dist_{point} + vehicle_radius
         Near2Point.clear();
         for (int k = 0; k < PointsOfInterest[wp_ind].size(); k++) {

            // --- Finds the closest path node to the point of interest --- //
            double infi         = std::numeric_limits<double>::infinity();
            vec3   NearestPoint = vec3(infi, infi, infi);
            for (int i = 1; i < N; i++) {
               auto vec = vertcat(x(i), y(i), z(i));

               if ((PoiOfInt[wp_ind][k] - path[wp_ind][i + ind]).norm() < (PoiOfInt[wp_ind][k] - NearestPoint).norm()) {
                  Nearest2Point = vec;
                  NearestPoint  = path[wp_ind][i + ind];
               }
            }

            // --- Sets the constraint --- //
            if (NearestPoint != path[wp_ind][N - 1 + ind]) {
               Near2Point.push_back(NearestPoint);

               opti.subject_to(Vehicle_radius < norm_2(Nearest2Point - (PointsOfInterest[wp_ind][k])) < observation_dist + Vehicle_radius);
            }
            
         }


         solver_options["qpsol"] = "qrqp";                 // Defines the QP solver used in SQP method
         //--------- SOLVER PRINT ----------
         solver_options["qpsol_options.print_iter"]   = 0; // Iteration print for qpsol
         solver_options["qpsol_options.print_header"] = 0; // Heading print for qpsol
         solver_options["print_header"]               = 0; // Heading print for solver
         solver_options["print_iteration"]            = 0; // Iteration print for solver
         solver_options["print_time"]                 = 0; // Optimation time print
         //-------------------------------------

         opti.solver("sqpmethod", solver_options);         // Initializes the Solver


         //----------- Initial Guess -------------
         for (int i = 0; i < N; i++) 
         {
            opti.set_initial(x(i), path[wp_ind][i + ind][0]); // x_0 = Global Path x
            opti.set_initial(y(i), path[wp_ind][i + ind][1]); // y_0 = Global Path y
            opti.set_initial(z(i), path[wp_ind][i + ind][2]); // z_0 = Global Path z
         }
         

         auto sol = opti.solve(); // Solves the Optimization problem


         // --- Extracts the solution ---//
         for (int n = 0; n < N; n++) {
            vec3 newNode(double(sol.value(x(n))), double(sol.value(y(n))), double(sol.value(z(n))));
            m_Path.at(n + ind) = newNode;
         }

         now = clock();
      }
      
      // --- Updates the Path --- //
      UpdatePath(NED);
   }


   const double* NMPC::OutNextNode(const double T, const double* const X)
	{
		return m_NextNode.data();
	}
   const double* NMPC::OutWaypoint(const double T, const double* const X, const int index)
   {
      return m_Waypoints[index].data();
   }
	const double* NMPC::OutObstacle(const double T, const double* const X, const int index)
	{
           return 0;
	}

	const double* NMPC::OutNode(const double T, const double* const X, int index)
	{
		if (index >= m_Path.size())
		{
              m_OutNodes[index](0) = 0; //nan("0");
              m_OutNodes[index](1) = 0; //nan("0");
              m_OutNodes[index](2) = 0; //nan("0");
		}
		else
		{
			m_OutNodes[index] = m_Path[index];		
		}
		return m_OutNodes[index].data();
	}

   void NMPC::FinalSetup(double T, const double* X, ISimObjectCreator* creator) 
   { 
      m_FhSimLogger = reinterpret_cast<CPrintDuringExec*>(creator->GetSharedResource("CPrintDuringExec"));
   }

   #ifdef FH_VISUALIZATION
   void NMPC::RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator)
   {
      // Render the path?
      creator->GetBoolParam("Render", &m_Render, true);

      if (m_Render) {

         m_OgreRoot = ogreRoot;
         m_SceneMgr = m_OgreRoot->getSceneManager("main");

         m_RenderDynamicLine     = new CDynamicLines(Ogre::RenderOperation::OT_LINE_LIST);
         m_RenderDynamicLineNode = m_SceneMgr->getRootSceneNode()->createChildSceneNode(m_SimObjectName + "Path");
         m_RenderDynamicLineNode->attachObject(m_RenderDynamicLine);

         // Render obstacles
         std::string material;
         creator->GetStringParam("RenderObstacleMaterial", material, "red2");
         for (int i = 0; i < m_ObstacleNumber; i++) {
            //const double* obstaclePos = m_InObstaclePos[i]->GetPortValue(T, X);
            m_RenderObstacles.push_back(m_SceneMgr->getRootSceneNode()->createChildSceneNode(m_SimObjectName + "Obstacle" + sfh::text::Int2Str(i)));
            m_RenderEntities.push_back(m_SceneMgr->createEntity(m_SimObjectName + "Obstacle" + sfh::text::Int2Str(i), "fhSphere.mesh"));
            m_RenderObstacles.at(i)->attachObject(m_RenderEntities.at(i));
            m_RenderEntities.at(i)->setMaterialName(material);
            m_RenderObstacles.at(i)->scale(2*obs_radius_init[i], 2*obs_radius_init[i], 2*obs_radius_init[i]); // float(2)
         }
      }
   }

   void NMPC::RenderUpdate(const double T, const double* const X)
   {
      if (m_Render) {
         // If the number of points in the previously rendered path differs from the number of the points in the updated path,
         // draw the rendered path from scratch.
         if (m_RenderDynamicLine->getNumPoints() != m_Path.size()) {
            m_RenderDynamicLine->clear();
            for (std::vector<vec3>::size_type i = 0; i < m_Path.size(); i++) {
               Ogre::Vector3 point = Ogre::Vector3(m_Path[i](0), m_Path[i](1), m_Path[i](2)); // Add the i-th node to the rendered line
               m_RenderDynamicLine->addPoint(point);
               if (i < m_Path.size() - 1) // If not at end of path, also add the next one
               {
                  point = Ogre::Vector3(m_Path[i + 1](0), m_Path[i + 1](1), m_Path[i + 1](2));
                  m_RenderDynamicLine->addPoint(point);
               }
            }
         } else // Else, iterate and update the position of each node
         {
            for (std::vector<vec3>::size_type i = 0; i < m_Path.size(); i++) {
               Ogre::Vector3 point = Ogre::Vector3(m_Path[i](0), m_Path[i](1), m_Path[i](2));
               m_RenderDynamicLine->setPoint(i, point);
               if (i < m_Path.size() - 1) {
                  point = Ogre::Vector3(m_Path[i + 1](0), m_Path[i + 1](1), m_Path[i + 1](2));
                  m_RenderDynamicLine->setPoint(i + 1, point);
               }
            }
         }
         m_RenderDynamicLine->update(); // Finally, update the path

         // Obstacles
         for (int i = 0; i < m_ObstacleNumber; i++) {
            const double* Obs = m_InObstaclePos[i]->GetPortValue(T, X);
            m_RenderObstacles.at(i)->setPosition(Obs[0], Obs[1], Obs[2]);
         }
      }
   }
#endif
   }
