#ifndef NMPC_H
#define NMPC_H

#include "SimObject.h"
#include <CPrintDuringExec.h>
//#include <gsl/gsl_vector.h>
//#include <gsl/gsl_multimin.h>
#include <casadi/casadi.hpp>
#include "HelpFunctions.h"

#ifdef FH_VISUALIZATION
#include <Ogre.h>
#include <sfh/ogre/CDynamicLines.h>
#endif

namespace AquaRobotics {

	class NMPC : public SimObject
	{
	public:
		/// The constructor reads the input parameters, sets up input and output ports.
		NMPC(const string& simObjectName, ISimObjectCreator* const creator);

		/// Destructor
                ~NMPC();

      void InitializePath();

      void UpdateInitialPath();

      void UpdatePath(vec3 NED);

      void UpdatePathNodes();

      void Calc_N();

      void Dynamic2Static(vec3 NED);

      casadi::MX ObjectFunction(casadi::MX x, casadi::MX y, casadi::MX z);

      void UpdatePositions(const double T, const double* const X);

		void PreOdeFcn(const double T, const double *const X, IStateUpdater* updater);

		/// Returns object derivatives as a function of time, states and input ports.
		void OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep);

      virtual void FinalSetup(double T, const double* X, ISimObjectCreator* creator);

#ifdef FH_VISUALIZATION
      /**
			\brief Initializes visualization resources at the simulation start-up.
		*/
      virtual void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator);
      /**
			\brief Updates visualization resources to reflect current state. Called once every timestep
		*/
      virtual void RenderUpdate(const double T, const double* const X);
#endif

	private:
      CPrintDuringExec*         m_FhSimLogger;
      ISignalPort*              m_InVehiclePos;
      ISignalPort*              m_InVehicleVel;
      ISignalPort*              m_InVehicleAcc;
      ISignalPort**             m_InObstaclePos;
      ISignalPort**             m_InObstacleVel;
      ISignalPort*              m_InWaypoint; 
      vec3                      m_NextNode;
      vec3                      VehicleVelocity;
      vec3                      VehicleForce;
      vec3                      hold_vec;
      std::vector<vec3>         Near2Point;
      casadi::MX                Nearest2Point;
      std::vector<vec3>         m_OutNodes;
      int                       m_ObstacleNumber;
      int                       m_WaypointNumber;
      double                    m_ObstacleRad;
      const double*             ROVPos;
      int                       N;
      int                       N_min;
      int                       n;
      int                       ang_count = 0;
      int                       hold = 0;
      int                       hold_obs = 0;
      int                       n_new=0;
      int                       n_old=0;
      int                       path_prev_index = 0;
      int                       path_index = 0;
      int                       n_prime;
      int                       prev_S = 0;
      int                       count = 1;
      int                       u = 0;
      int                       ind = 0;
      int                       N_init = 0;
      int                       extra_count = 0;
      int                       wp_ind      = 0;
      int                       inObs       = 0;
      int                       speed_count = 1;
      int                       diff = 0;
      double                    Vehicle_radius;
      double                    observation_dist;
      double                    Vehicle_mass;
      double                    time_st= 0.0;
      double                    ang_pathobs;
      double                    Obs_time = 0.0;
      float                     Update_frequency;
      float                     seg_len;
      int                       m_NodesToLog;
      clock_t                   now;
      clock_t                   Sample_Obs_time;
      float                     break_time;
      bool                     collision;
      std::vector<vec3>              path_wp;
      std::vector < std::vector<vec3>> PoiOfInt;
      std::vector < std::vector<casadi::MX>> PointsOfInterest;
      std::vector<std::vector<vec3>> path;
      casadi::Dict              solver_options;
      std::vector<casadi::MX>   obs;
      std::vector<double>       obs_radius;
      std::vector<double>       obs_radius_init;
      std::vector<double>       StateTime;
      std::vector<double>       runtime;
      std::vector<int>          horizons;
      std::vector<int>          collision_index;
      std::vector<double>       Obstacle_dist;
      std::vector<vec3>         m_Obs;
      std::vector<vec3>         m_Obs_speed;

      std::vector<vec3>         m_Obs_new;
      std::vector<vec3>         m_Obs_hold;
      std::vector<vec3>         m_Obs_collision_points;
      std::vector<vec3>         m_Obs_pos_new;
      std::vector<std::vector<vec3>> m_Obs_collision;
      std::vector<std::vector<vec3>> collisions;
      std::vector<vec3>         m_Obs_time_vec;
      std::vector<std::vector<vec3>> m_Obs_time_vec_all;
      
      std::vector<vec3>         m_Waypoints;
      std::vector<vec3>         m_Path;

      const double* OutNextNode(const double T, const double* const X);
      const double* OutNode(const double T, const double* const X, int index);
      const double* OutWaypoint(const double T, const double* const X, const int index);
      const double* OutObstacle(const double T, const double* const X, const int index);

#ifdef FH_VISUALIZATION
           Ogre::Root*                m_OgreRoot;              ///< The Ogre root object.
           Ogre::SceneManager*        m_SceneMgr;              ///< A pointer to the scene manager.
           bool                       m_Render;                ///< Whether or not to render the path
           vector<Ogre::SceneNode*>   m_RenderObstacles;       ///< The node to which the rendered entity is attached.
           std::vector<Ogre::Entity*> m_RenderEntities;        ///< The entities to be rendered in Ogre.
           CDynamicLines*             m_RenderDynamicLine;     ///< The path rendered as a dynamic line
           Ogre::SceneNode*           m_RenderDynamicLineNode; ///< The node which the dynamic line is attached
#endif
	};

}

#endif //NMPC_H
